@@ -50,6 +50,10 @@ static gboolean   xfce_appfinder_window_key_press_event               (GtkWidget
                                                                        GdkEventKey                 *event);
 static void       xfce_appfinder_window_set_padding                   (GtkWidget                   *entry,
                                                                        GtkWidget                   *align);
+static gboolean   xfce_appfinder_window_completion_match_func         (GtkEntryCompletion          *completion,
+                                                                       const gchar                 *key,
+                                                                       GtkTreeIter                 *iter,
+                                                                       gpointer                     data);
 static void       xfce_appfinder_window_entry_changed                 (XfceAppfinderWindow         *window);
 static void       xfce_appfinder_window_entry_activate                (GtkEditable                 *entry,
                                                                        XfceAppfinderWindow         *window);
@@ -69,6 +73,9 @@ static void       xfce_appfinder_window_drag_data_get                 (GtkWidget
                                                                        guint                        info,
                                                                        guint                        drag_time,
                                                                        XfceAppfinderWindow         *window);
+static gboolean   xfce_appfinder_window_treeview_key_press_event      (GtkWidget                   *widget,
+                                                                       GdkEventKey                 *event,
+                                                                       XfceAppfinderWindow         *window);
 static void       xfce_appfinder_window_category_changed              (GtkTreeSelection            *selection,
                                                                        XfceAppfinderWindow         *window);
 static void       xfce_appfinder_window_category_set_categories       (XfceAppfinderWindow         *window);
@@ -226,6 +233,7 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
 
   window->completion = completion = gtk_entry_completion_new ();
   gtk_entry_completion_set_model (completion, GTK_TREE_MODEL (window->model));
+  gtk_entry_completion_set_match_func (completion, xfce_appfinder_window_completion_match_func, window, NULL);
   gtk_entry_completion_set_text_column (completion, XFCE_APPFINDER_MODEL_COLUMN_COMMAND);
   gtk_entry_completion_set_popup_completion (completion, TRUE);
   gtk_entry_completion_set_popup_single_match (completion, FALSE);
@@ -251,6 +259,7 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (sidepane), FALSE);
   gtk_tree_view_set_enable_search (GTK_TREE_VIEW (sidepane), FALSE);
   g_signal_connect_swapped (G_OBJECT (sidepane), "start-interactive-search", G_CALLBACK (gtk_widget_grab_focus), entry);
+  g_signal_connect (G_OBJECT (sidepane), "key-press-event", G_CALLBACK (xfce_appfinder_window_treeview_key_press_event), window);
   gtk_tree_view_set_row_separator_func (GTK_TREE_VIEW (sidepane),
       xfce_appfinder_category_model_row_separator_func, NULL, NULL);
   gtk_container_add (GTK_CONTAINER (scroll), sidepane);
@@ -295,6 +304,7 @@ xfce_appfinder_window_init (XfceAppfinderWindow *window)
   gtk_drag_source_set (treeview, GDK_BUTTON1_MASK, target_list, G_N_ELEMENTS (target_list), GDK_ACTION_COPY);
   g_signal_connect (G_OBJECT (treeview), "drag-begin", G_CALLBACK (xfce_appfinder_window_drag_begin), window);
   g_signal_connect (G_OBJECT (treeview), "drag-data-get", G_CALLBACK (xfce_appfinder_window_drag_data_get), window);
+  g_signal_connect (G_OBJECT (treeview), "key-press-event", G_CALLBACK (xfce_appfinder_window_treeview_key_press_event), window);
   gtk_container_add (GTK_CONTAINER (scroll), treeview);
   gtk_widget_show (treeview);
   g_object_unref (G_OBJECT (filter_model));
@@ -454,6 +464,29 @@ xfce_appfinder_window_set_padding (GtkWidget *entry,
 
 
 
+static gboolean
+xfce_appfinder_window_completion_match_func (GtkEntryCompletion *completion,
+                                             const gchar        *key,
+                                             GtkTreeIter        *iter,
+                                             gpointer            data)
+{
+  const gchar *text;
+
+  XfceAppfinderWindow *window = XFCE_APPFINDER_WINDOW (data);
+
+  appfinder_return_val_if_fail (GTK_IS_ENTRY_COMPLETION (completion), FALSE);
+  appfinder_return_val_if_fail (XFCE_IS_APPFINDER_WINDOW (data), FALSE);
+  appfinder_return_val_if_fail (GTK_TREE_MODEL (window->model)
+      == gtk_entry_completion_get_model (completion), FALSE);
+
+  /* don't use the casefolded key generated by gtk */
+  text = gtk_entry_get_text (GTK_ENTRY (window->entry));
+
+  return xfce_appfinder_model_get_visible_command (window->model, iter, text);
+}
+
+
+
 static gboolean
 xfce_appfinder_window_entry_changed_idle (gpointer data)
 {
@@ -518,7 +551,7 @@ xfce_appfinder_window_entry_changed (XfceAppfinderWindow *window)
     g_source_remove (window->idle_entry_changed_id);
 
   window->idle_entry_changed_id =
-      g_idle_add_full (G_PRIORITY_LOW, xfce_appfinder_window_entry_changed_idle,
+      g_idle_add_full (G_PRIORITY_DEFAULT, xfce_appfinder_window_entry_changed_idle,
                        window, xfce_appfinder_window_entry_changed_idle_destroyed);
 }
 
@@ -555,9 +588,9 @@ xfce_appfinder_window_entry_key_press_event (GtkWidget           *entry,
 {
   gboolean expand, is_expanded;
 
-  if (event->keyval == GDK_KEY_Up || event->keyval == GDK_KEY_Down)
+  if (event->keyval == GDK_Up || event->keyval == GDK_Down)
     {
-      expand = (event->keyval == GDK_KEY_Down);
+      expand = (event->keyval == GDK_Down);
       is_expanded = gtk_widget_get_visible (window->paned);
       if (is_expanded != expand)
         {
@@ -629,6 +662,33 @@ xfce_appfinder_window_drag_data_get (GtkWidget           *widget,
 
 
 
+static gboolean
+xfce_appfinder_window_treeview_key_press_event (GtkWidget           *widget,
+                                                GdkEventKey         *event,
+                                                XfceAppfinderWindow *window)
+{
+  if (widget == window->treeview)
+    {
+      if (event->keyval == GDK_Left)
+        {
+          gtk_widget_grab_focus (window->sidepane);
+          return TRUE;
+        }
+    }
+  else if (widget == window->sidepane)
+    {
+      if (event->keyval == GDK_Right)
+        {
+          gtk_widget_grab_focus (window->treeview);
+          return TRUE;
+        }
+    }
+
+  return FALSE;
+}
+
+
+
 static void
 xfce_appfinder_window_entry_icon_released (GtkEntry             *entry,
                                            GtkEntryIconPosition  icon_pos,
@@ -803,6 +863,7 @@ static gboolean
 xfce_appfinder_window_execute_command (const gchar          *text,
                                        GdkScreen            *screen,
                                        XfceAppfinderWindow  *window,
+                                       gboolean             *save_cmd,
                                        GError              **error)
 {
   gboolean  succeed = FALSE;
@@ -819,7 +880,7 @@ xfce_appfinder_window_execute_command (const gchar          *text,
     window->actions = xfce_appfinder_actions_get ();
 
   /* try to match a custom action */
-  action_cmd = xfce_appfinder_actions_execute (window->actions, text, error);
+  action_cmd = xfce_appfinder_actions_execute (window->actions, text, save_cmd, error);
   if (*error != NULL)
     return FALSE;
   else if (action_cmd != NULL)
@@ -855,6 +916,7 @@ xfce_appfinder_window_execute (XfceAppfinderWindow *window)
   const gchar      *text;
   gchar            *cmd = NULL;
   gboolean          regular_command = FALSE;
+  gboolean          save_cmd;
 
   if (!gtk_widget_get_sensitive (window->button_launch))
     return;
@@ -871,7 +933,7 @@ xfce_appfinder_window_execute (XfceAppfinderWindow *window)
           if (!result && regular_command)
             {
               gtk_tree_model_get (model, &iter, XFCE_APPFINDER_MODEL_COLUMN_COMMAND, &cmd, -1);
-              result = xfce_appfinder_window_execute_command (cmd, screen, window, &error);
+              result = xfce_appfinder_window_execute_command (cmd, screen, window, NULL, &error);
               g_free (cmd);
             }
         }
@@ -879,8 +941,15 @@ xfce_appfinder_window_execute (XfceAppfinderWindow *window)
   else
     {
       text = gtk_entry_get_text (GTK_ENTRY (window->entry));
-      if (xfce_appfinder_window_execute_command (text, screen, window, &error))
-        result = xfce_appfinder_model_save_command (window->model, text, &error);
+      save_cmd = TRUE;
+
+      if (xfce_appfinder_window_execute_command (text, screen, window, &save_cmd, &error))
+        {
+          if (save_cmd)
+            result = xfce_appfinder_model_save_command (window->model, text, &error);
+          else
+            result = TRUE;
+        }
     }
 
   gtk_entry_set_icon_from_stock (GTK_ENTRY (window->entry), GTK_ENTRY_ICON_PRIMARY,
